#include "wino_macro.h"
#include "../software/param.h"
#include <ap_int.h>
#include <cstdlib>




#ifndef _WINO_IO_HPP_
#define _WINO_IO_HPP_

template<int port_idx>
void load_input_row_from_ddr(
		ap_uint<128>* DDR_port,
		ap_uint<16> input_buffer[INBUFFER_WIDTH][INPUT_BUFFER_DEPTH],
		ap_uint<16> inheight,
		ap_uint<16> inwidth,
		ap_uint<16> inwidth_align8,
		ap_uint<16> indepth_align8,
		ap_uint<16> group_indepth_x_inwidth_align8_by8,
		ap_uint<16> group_indepth_offset_x_inwidth_align8_by8,
		ap_uint<16> inwidth_ceildiv_inbufferwidth,
		ap_uint<16> buffer_address_mid_increment_step,
		ap_uint<16> input_load_burst_length,
		ap_uint<1> row_address_bitnumber_flag,
		ap_uint<16> row_idx,
		ap_int<16> row_load_bound)
{
	if(row_idx >= inheight || row_idx >= row_load_bound) return;



	#if DEBUG_FILE_PRINT
		printf("Loading row_idx %d, %d\n", (int) row_idx, port_idx);
	#endif

	#pragma HLS interface m_axi port=DDR_port depth=200704
	#pragma HLS array_partition variable = input_buffer dim=1 complete

	ap_uint<16> register_matrix[8][8];
	#pragma HLS array_partition variable=register_matrix dim=1 complete
	#pragma HLS array_partition variable=register_matrix dim=2 complete

	ap_uint<1> row_pingpong_bit1 = row_idx[INBUFFER_HEIGHT_BITWIDTH];
	ap_uint<2> row_pingpong_bit2 = row_idx.range(INBUFFER_HEIGHT_BITWIDTH+1, INBUFFER_HEIGHT_BITWIDTH);


	ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid=0;
	ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_offset=0;


	printf("row_address_bitnumber_flag %d\n", (int) row_address_bitnumber_flag);
	#if DEBUG_FILE_PRINT
		if(row_address_bitnumber_flag)
		{
			if( inwidth_ceildiv_inbufferwidth*indepth_align8 > (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 1)) )
			{
				printf("ERROR: in sufficient IN BUFFER DEPTH, expected depth=[%d x %d], reald depth [%d]\n", (int) inwidth_ceildiv_inbufferwidth, (int) indepth_align8, (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 1)) );
				exit(-1);
			}
		}
		else
		{
			if( inwidth_ceildiv_inbufferwidth*indepth_align8 > (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 2)) )
			{
				printf("ERROR: in sufficient IN BUFFER DEPTH, expected depth=[%d x %d], reald depth [%d]\n", (int) inwidth_ceildiv_inbufferwidth, (int) indepth_align8, (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 2)) );
				exit(-1);
			}
		}
		
	#endif



	ap_uint<INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mini_tile=0;

	ap_uint<3> depth_idx8=0;

	ap_uint<8> read_inwidth_counter=1;
	ap_uint<8> write_inwidth_counter=1;

	ap_uint<32> ddr_address_offset = row_idx * group_indepth_x_inwidth_align8_by8 + group_indepth_offset_x_inwidth_align8_by8;

	ap_uint<128>* DDR_offset = DDR_port + ddr_address_offset;
	
	ap_uint<1> left_down_flag=1;

	ap_uint<INBUFFER_WIDTH_BITWIDTH-3> bank_split_idx=0;

	for(ap_uint<16> cycle=0;cycle<input_load_burst_length;cycle++)
	{
		#pragma HLS pipeline
		ap_uint<128> DDR_data;

		DDR_data = DDR_offset[cycle];
	
		ap_uint<16> DDR_data_divided[8];
		#pragma HLS array_partition variable = DDR_data_divided complete

		for(int i=0;i<8;i++)
		{
		#pragma HLS unroll
			DDR_data_divided[i]=DDR_data.range(i*16+15,i*16);
		}

		if(read_inwidth_counter == inwidth_align8)
		{
			read_inwidth_counter=1;
		}
		else
		{
			read_inwidth_counter++;
		}
		

		


		ap_uint<16> bank_data[8];
		#pragma HLS array_partition variable = bank_data complete

		if(left_down_flag)
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[i][0];
			}
			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[j][i]=register_matrix[j][i+1];
				}
			}
			for(int i=0;i<8;i++)
			{ 
			#pragma HLS unroll
				register_matrix[i][7]=DDR_data_divided[i];
			}
		}
		else
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[0][i];
			}

			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[i][j]=register_matrix[i+1][j];
				}
			}
			for(int i=0;i<8;i++)
			{ 
			#pragma HLS unroll
				register_matrix[7][i]=DDR_data_divided[i];
			}
		}
		if(depth_idx8==7)
			left_down_flag=~left_down_flag;





		if(cycle>=8)
		{
			ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH> buffer_addr; 
			ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit1_offset;
			ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 2 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit2_offset;
			
			buffer_address_mid_rowbit1_offset=buffer_address_mid_offset;
			buffer_address_mid_rowbit2_offset=buffer_address_mid_offset;

			if(row_address_bitnumber_flag)
			{
				buffer_addr=(row_pingpong_bit1,buffer_address_mid_rowbit1_offset,buffer_address_mini_tile);
			}
			else
			{
				buffer_addr=(row_pingpong_bit2,buffer_address_mid_rowbit2_offset,buffer_address_mini_tile);
			}
			

			input_buffer[bank_split_idx*8+0][buffer_addr]=bank_data[0];
			input_buffer[bank_split_idx*8+1][buffer_addr]=bank_data[1];
			input_buffer[bank_split_idx*8+2][buffer_addr]=bank_data[2];
			input_buffer[bank_split_idx*8+3][buffer_addr]=bank_data[3];
			input_buffer[bank_split_idx*8+4][buffer_addr]=bank_data[4];
			input_buffer[bank_split_idx*8+5][buffer_addr]=bank_data[5];
			input_buffer[bank_split_idx*8+6][buffer_addr]=bank_data[6];
			input_buffer[bank_split_idx*8+7][buffer_addr]=bank_data[7];

			if(write_inwidth_counter == inwidth_align8)
			{
				buffer_address_mid += buffer_address_mid_increment_step;
			}
			else if( write_inwidth_counter.range(INBUFFER_WIDTH_BITWIDTH-1,0) == 0 )
			{
				buffer_address_mid++;
			}

			if( write_inwidth_counter.range(2,0) == 0)
			{
				buffer_address_mid_offset = buffer_address_mid;
			}
			else if(buffer_address_mini_tile == INDEPTH_MINITILE_SIZE-1)
			{
				buffer_address_mid_offset += inwidth_ceildiv_inbufferwidth;
			}

			buffer_address_mini_tile++;

			if(write_inwidth_counter == inwidth_align8)
			{
				bank_split_idx=0;
			}
			else if(write_inwidth_counter.range(2,0)==0)
			{
				bank_split_idx++;
			}
				
			if(write_inwidth_counter == inwidth_align8)
			{
				write_inwidth_counter=1;
			}
			else
			{
				write_inwidth_counter++;
			}
		}
		depth_idx8++;

	}

	for(ap_uint<16> cycle=0;cycle<8;cycle++)
	{
		#pragma HLS pipeline

		ap_uint<16> bank_data[8];
		#pragma HLS array_partition variable = bank_data complete

		if(left_down_flag)
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[i][0];
			}
			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[j][i]=register_matrix[j][i+1];
				}
			}
		}
		else
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[0][i];
			}

			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[i][j]=register_matrix[i+1][j];
				}
			}
		}
			
		ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH> buffer_addr; 
		ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit1_offset;
		ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 2 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit2_offset;
		
		buffer_address_mid_rowbit1_offset=buffer_address_mid_offset;
		buffer_address_mid_rowbit2_offset=buffer_address_mid_offset;

		if(row_address_bitnumber_flag)
		{
			buffer_addr=(row_pingpong_bit1,buffer_address_mid_rowbit1_offset,buffer_address_mini_tile);
		}
		else
		{
			buffer_addr=(row_pingpong_bit2,buffer_address_mid_rowbit2_offset,buffer_address_mini_tile);
		}
		input_buffer[bank_split_idx*8+0][buffer_addr]=bank_data[0];
		input_buffer[bank_split_idx*8+1][buffer_addr]=bank_data[1];
		input_buffer[bank_split_idx*8+2][buffer_addr]=bank_data[2];
		input_buffer[bank_split_idx*8+3][buffer_addr]=bank_data[3];
		input_buffer[bank_split_idx*8+4][buffer_addr]=bank_data[4];
		input_buffer[bank_split_idx*8+5][buffer_addr]=bank_data[5];
		input_buffer[bank_split_idx*8+6][buffer_addr]=bank_data[6];
		input_buffer[bank_split_idx*8+7][buffer_addr]=bank_data[7];

 		if(buffer_address_mini_tile == INDEPTH_MINITILE_SIZE-1)
		{
			buffer_address_mid_offset += inwidth_ceildiv_inbufferwidth;
		}
		buffer_address_mini_tile++;
	}
}




//template<int DUMMY>
void load_input_rowtile_from_ddr(
		ap_uint<128>* DDR_port1,
        ap_uint<128>* DDR_port2,
        ap_uint<128>* DDR_port3,
        ap_uint<128>* DDR_port4,
		ap_uint<16> input_buffer[INBUFFER_HEIGHT][INBUFFER_WIDTH][INPUT_BUFFER_DEPTH],
		ap_uint<16> inheight,
		ap_uint<16> inwidth,
		ap_uint<16> stride,
        ap_uint<16> pad_size,
		ap_uint<16> inwidth_align8,
		ap_uint<16> indepth_align8,
		ap_uint<16> group_indepth_x_inwidth_align8_by8,
		ap_uint<16> group_indepth_offset_x_inwidth_align8_by8,
		ap_uint<16> inwidth_ceildiv_inbufferwidth,
		ap_uint<16> buffer_address_mid_increment_step,
		ap_uint<16> input_load_burst_length,
		ap_uint<1>  row_address_bitnumber_flag,
		ap_uint<16> out_rowstep,
		ap_uint<16> next_start_output_row,
        ap_uint<1> first_flag
)
{
	#pragma HLS array_partition variable =input_buffer dim=1 complete
	#pragma HLS array_partition variable =input_buffer dim=2 complete
    static ap_uint<16> loaded_input_row_number;
    if(first_flag)
        loaded_input_row_number = 0;
    
    ap_int<16> required_loaded_input_row_number = (next_start_output_row+out_rowstep)*stride + pad_size;

    while( required_loaded_input_row_number > loaded_input_row_number)
    {
		ap_uint<16> row_idx[4];
		#pragma HLS array_partition variable=row_idx complete
		ap_uint<2> row_idx_sel[4];
		#pragma HLS array_partition variable=row_idx_sel complete

		ap_uint<16> row_idx_selected[4];
		#pragma HLS array_partition variable=row_idx complete

		for(int i=0;i<4;i++)
		{
			#pragma hls unroll 
			row_idx[i]=loaded_input_row_number+i;
			row_idx_sel[i]=i-loaded_input_row_number.range(1,0);
		}

		for(int i=0;i<4;i++)
		{
			#pragma hls unroll 
			row_idx_selected[i]=row_idx[row_idx_sel[i]];
		}

		printf("row_idx[%d %d %d %d]\n",(int) row_idx_selected[0],(int) row_idx_selected[1],(int) row_idx_selected[2],(int) row_idx_selected[3] );


		if( row_idx_selected[0]%INBUFFER_HEIGHT/4 == 0)
		{

			load_input_row_from_ddr<0>(
				DDR_port1,
				input_buffer[0],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[0],
				required_loaded_input_row_number
			);
		}
		else
		{
			load_input_row_from_ddr<0>(
				DDR_port1,
				input_buffer[4],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[0],
				required_loaded_input_row_number
			);
		}
	
		if( row_idx_selected[1]%INBUFFER_HEIGHT/4 == 0)
		{

			load_input_row_from_ddr<1>(
				DDR_port2,
				input_buffer[1],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[1],
				required_loaded_input_row_number
			);
		}
		else
		{
			load_input_row_from_ddr<1>(
				DDR_port2,
				input_buffer[5],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[1],
				required_loaded_input_row_number
			);
		}

		if( row_idx_selected[2]%INBUFFER_HEIGHT/4 == 0)
		{

			load_input_row_from_ddr<2>(
				DDR_port3,
				input_buffer[2],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[2],
				required_loaded_input_row_number
			);
		}
		else
		{
			load_input_row_from_ddr<2>(
				DDR_port3,
				input_buffer[6],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[2],
				required_loaded_input_row_number
			);
		}


		if( row_idx_selected[3]%INBUFFER_HEIGHT/4 == 0)
		{

			load_input_row_from_ddr<3>(
				DDR_port4,
				input_buffer[3],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[3],
				required_loaded_input_row_number
			);
		}
		else
		{
			load_input_row_from_ddr<3>(
				DDR_port4,
				input_buffer[7],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
				row_idx_selected[3],
				required_loaded_input_row_number
			);
		}

        loaded_input_row_number+=4;
    }
	loaded_input_row_number=required_loaded_input_row_number;
	#if DEBUG_FILE_PRINT
	fflush(stdout);
	#endif
}


template<int port_idx>
void write_output_to_DDR_onerow(
	ap_uint<ODDR_WIDTH*BATCH_SIZE*8>* out_DDR,
	ap_uint<OUT_WIDTH*2> out_buffer[WINO_OUT_SIZE][OUTDEPTH_MINITILE_SIZE][WINO_WIDTH][OUTPUT_BUFFER_DEPTH],
	ap_uint<16> outheight,
	ap_uint<16> outwidth_align8,
	ap_uint<4> wino_output_tile_size,
	ap_uint<4> wino_col_pix_upper_bound,
	ap_uint<8> wino_tile_number_rowcol,
	ap_uint<24> output_burst_length,
	ap_uint<OUTPUT_BUFFER_DEPTH> rowtile_baseaddr,
	ap_uint<16> start_row_idx,
	#if DEBUG_FILE_PRINT
	ConvDesc_t conv_desc
	#endif
)
{
	#if WINO_OUT_SIZE == 4
	
	if(port_idx >= wino_output_tile_size || start_row_idx >= outheight)
	{
		return;
	}
	#else
	#endif
	ap_uint<WINO_WIDTH_BITWIDTH> wino_widthtile_idx=0;
	
	ap_uint<4> wino_col_pix_idx=0;



	ap_uint<2> depthtile_idx=0;

	ap_uint<16> col_idx_cnt=1;

	ap_uint<OUTPUT_BUFFER_DEPTH_BITWIDTH> coltile_baseaddr=0;
	ap_uint<OUTPUT_BUFFER_DEPTH_BITWIDTH> depthtile_baseaddr0=0;

	#if WEIGHT_FEED_NUMBER_PER_PORT == 1
	ap_uint<OUTPUT_BUFFER_DEPTH_BITWIDTH> depthtile_baseaddr1=wino_tile_number_rowcol;
	#endif



	#if WEIGHT_FEED_NUMBER_PER_PORT == 1
	ap_uint<8> depthtile_baseaddr_increment_step=wino_tile_number_rowcol*2;
	#else
	ap_uint<8> depthtile_baseaddr_increment_step=wino_tile_number_rowcol;
	#endif


	// printf("write_length %d\n", (int) conv_desc.output_burst_length);
	for(ap_uint<24> cycle=0;cycle < output_burst_length;cycle++)
	{
		#if DEBUG_FILE_PRINT
			int col_idx=cycle%conv_desc.outdepth_align8;
			int depth_idx_by8=cycle/conv_desc.outwidth_align8;
		#endif
		
		ap_uint<OUT_WIDTH*BATCH_SIZE> outbuffer_data_depth0[OUTDEPTH_MINITILE_SIZE];
		#pragma HLS array_partition variable =outbuffer_data_depth0 complete

		#if WEIGHT_FEED_NUMBER_PER_PORT == 1
		ap_int<OUT_WIDTH*BATCH_SIZE> outbuffer_data_depth1[OUTDEPTH_MINITILE_SIZE];
		#pragma HLS array_partition variable =outbuffer_data_depth1 complete
		#endif

		ap_uint<OUTPUT_BUFFER_DEPTH_BITWIDTH> buffer_address0;
		buffer_address0=depthtile_baseaddr0+rowtile_baseaddr+coltile_baseaddr;

		#if WEIGHT_FEED_NUMBER_PER_PORT == 1
			ap_uint<OUTPUT_BUFFER_DEPTH_BITWIDTH> buffer_address1;
			buffer_address1=depthtile_baseaddr1+rowtile_baseaddr+coltile_baseaddr;
		#endif

		// printf("outbuffer add %d %d %d %d\n",(int) coltile_baseaddr,(int)wino_widthtile_idx,(int) wino_col_pix_idx, (int) wino_col_pix_upper_bound );

		for(int b=0;b<BATCH_SIZE;b++)
		{
			#pragma HLS unroll
			for(int od=0;od<OUTDEPTH_MINITILE_SIZE;od++)
			{
				#pragma HLS unroll
				outbuffer_data_depth0[od]=out_buffer[wino_col_pix_idx][od][wino_widthtile_idx][buffer_address0];
				
				#if WEIGHT_FEED_NUMBER_PER_PORT == 1
				outbuffer_data_depth1[od]=out_buffer[wino_col_pix_idx][od][wino_widthtile_idx][buffer_address1];
				#endif
			}
		}
		ap_int<ODDR_WIDTH>  outmem_data[8][BATCH_SIZE];

		#if WEIGHT_FEED_NUMBER_PER_PORT ==1
		
			for(int i=0;i<4;i++)
			{
				#pragma HLS unroll
				for(int b=0;b<BATCH_SIZE;b++)
				{	
					#pragma HLS unroll
					outmem_data[i][b]= (outbuffer_data_depth0[i].range(b*OUT_WIDTH+OUT_WIDTH-1,b*OUT_WIDTH))>>(OUT_WIDTH- ODDR_WIDTH);
					outmem_data[i+4][b]= (outbuffer_data_depth1[i].range(b*OUT_WIDTH+OUT_WIDTH-1,b*OUT_WIDTH))>>(OUT_WIDTH- ODDR_WIDTH);
				}
			}
		#else
			for(int i=0;i<8;i++)
			{
				#pragma HLS unroll
				for(int b=0;b<BATCH_SIZE;b++)
				{	
					#pragma HLS unroll
					outmem_data[i][b]= (outbuffer_data_depth0[i].range(b*OUT_WIDTH+OUT_WIDTH-1,b*OUT_WIDTH))>>(OUT_WIDTH- ODDR_WIDTH);
				}
			}
		#endif

		ap_int<ODDR_WIDTH*8*BATCH_SIZE> write_back_value;

		for(int i=0;i<8;i++)
		{
			#pragma HLS unroll
			for(int b=0;b<BATCH_SIZE;b++)
			{
				#pragma HLS unroll
				write_back_value.range( (i*BATCH_SIZE+b)*ODDR_WIDTH+ODDR_WIDTH-1, (i*BATCH_SIZE+b)*ODDR_WIDTH )=outmem_data[i][b];
			}
		}
		out_DDR[cycle]=write_back_value;

		if(col_idx_cnt==outwidth_align8 )
		{
			depthtile_baseaddr0+=depthtile_baseaddr_increment_step;
			#if WEIGHT_FEED_NUMBER_PER_PORT == 1
			depthtile_baseaddr1+=depthtile_baseaddr_increment_step;
			#endif
		}




		if(col_idx_cnt==outwidth_align8 )
		{
			coltile_baseaddr=0;
		}
		else if( wino_col_pix_idx == wino_col_pix_upper_bound  && wino_widthtile_idx == WINO_WIDTH-1  )
		{
			coltile_baseaddr++;
		}


		if( wino_col_pix_idx == wino_col_pix_upper_bound  && wino_widthtile_idx == WINO_WIDTH-1 || col_idx_cnt==outwidth_align8)
		{
			wino_widthtile_idx=0;
		}
		else if( wino_col_pix_idx == wino_col_pix_upper_bound )
		{
			wino_widthtile_idx++;
		}


		if(wino_col_pix_idx == wino_col_pix_upper_bound || col_idx_cnt==outwidth_align8)
		{
			wino_col_pix_idx=0;
		}
		else
		{
			wino_col_pix_idx++;
		}

		if(col_idx_cnt==outwidth_align8)
		{
			col_idx_cnt=1;
		}
		else
		{
			col_idx_cnt++;
		}
	}
}



void write_output_to_DDR(
		ap_uint<ODDR_WIDTH*BATCH_SIZE*8>* out_DDR0,
		ap_uint<ODDR_WIDTH*BATCH_SIZE*8>* out_DDR1,
		ap_uint<ODDR_WIDTH*BATCH_SIZE*8>* out_DDR2,
		ap_uint<ODDR_WIDTH*BATCH_SIZE*8>* out_DDR3,
		ap_uint<OUT_WIDTH*2> out_buffer[WINO_OUT_SIZE][WINO_OUT_SIZE][OUTDEPTH_MINITILE_SIZE][WINO_WIDTH][OUTPUT_BUFFER_DEPTH],
		ap_uint<16> outheight,
		ap_uint<16> outwidth_align8,
		ap_uint<4> wino_output_tile_size,
		ap_uint<4> wino_col_pix_upper_bound,
		ap_uint<8> wino_tile_number_rowcol,
		ap_uint<24> output_burst_length,
		ap_uint<16> start_row_idx,
		ap_uint<1> first_flag,
		#if DEBUG_FILE_PRINT
		ConvDesc_t conv_desc
		#endif
)
{

	static ap_uint<24> out_ddr_offset0;
	static ap_uint<24> out_ddr_offset1;
	static ap_uint<24> out_ddr_offset2;
	static ap_uint<24> out_ddr_offset3;


	ap_uint<16> outrow_idx=start_row_idx;
	if( first_flag)
	{
		out_ddr_offset0=0;
		out_ddr_offset1=conv_desc.output_burst_length;
		out_ddr_offset2=(conv_desc.output_burst_length)<<1;
		out_ddr_offset3=out_ddr_offset1+out_ddr_offset2;
	}

	ap_uint<16> rowtile_baseaddr=0;

	for(ap_uint<4> wino_rowtile_idx=0; wino_rowtile_idx<conv_desc.wino_tile_number_in_out_rowstep;wino_rowtile_idx++)
	{
		printf("out_ddr_offset0 %d\n",(int) out_ddr_offset0);
		printf("out_ddr_offset1 %d\n",(int) out_ddr_offset1);
		printf("out_ddr_offset2 %d\n",(int) out_ddr_offset2);
		printf("out_ddr_offset3 %d\n",(int) out_ddr_offset3);
		write_output_to_DDR_onerow<0>(
			out_DDR0+out_ddr_offset0,
			out_buffer[0],
			outheight,
			outwidth_align8,
			wino_output_tile_size,
			wino_col_pix_upper_bound,
			wino_tile_number_rowcol,
			output_burst_length,
			rowtile_baseaddr,
			outrow_idx+0,
			#if DEBUG_FILE_PRINT
			conv_desc
			#endif
			);
		write_output_to_DDR_onerow<1>(
			out_DDR1+out_ddr_offset1,
			out_buffer[1],
			outheight,
			outwidth_align8,
			wino_output_tile_size,
			wino_col_pix_upper_bound,
			wino_tile_number_rowcol,
			output_burst_length,
			rowtile_baseaddr,
			outrow_idx+1,
			#if DEBUG_FILE_PRINT
			conv_desc
			#endif
			);
		#if WINO_OUT_SIZE == 4
		write_output_to_DDR_onerow<2>(
			out_DDR2+out_ddr_offset2,
			out_buffer[2],
			outheight,
			outwidth_align8,
			wino_output_tile_size,
			wino_col_pix_upper_bound,
			wino_tile_number_rowcol,
			output_burst_length,
			rowtile_baseaddr,
			outrow_idx+2,
			#if DEBUG_FILE_PRINT
			conv_desc
			#endif
			);
		write_output_to_DDR_onerow<3>(
			out_DDR3+out_ddr_offset3,
			out_buffer[3],
			outheight,
			outwidth_align8,
			wino_output_tile_size,
			wino_col_pix_upper_bound,
			wino_tile_number_rowcol,
			output_burst_length,
			rowtile_baseaddr,
			outrow_idx+3,
			#if DEBUG_FILE_PRINT
			conv_desc
			#endif
			);
		#endif
		rowtile_baseaddr+=conv_desc.wino_tile_number_in_outwidth;
		out_ddr_offset0+=conv_desc.output_burst_length*conv_desc.wino_output_tile_size;
		out_ddr_offset1+=conv_desc.output_burst_length*conv_desc.wino_output_tile_size;
		out_ddr_offset2+=conv_desc.output_burst_length*conv_desc.wino_output_tile_size;
		out_ddr_offset3+=conv_desc.output_burst_length*conv_desc.wino_output_tile_size;
		outrow_idx+=conv_desc.wino_output_tile_size;
	}
}

#endif
